"""Context module

I really hate the ineloquence of passing vars, batch and run
as the context to the tasks. I definitely want a proper context
manager for bringing the contexts together into a mapping
"""
